import h5py
import numpy as np
import cv2
import math
import json
import sys
from scipy import interpolate
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame
from scipy.interpolate import UnivariateSpline
from numpy import linspace
# from kalmanFilter import kalman_xy

from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle
from zebrazoom.code.trackingFolder.trackingFunctions import distBetweenThetas
from zebrazoom.code.trackingFolder.trackingFunctions import assignValueIfBetweenRange

def dist2(v, w):
  return (v["x"] - w["x"])**2 + (v["y"] - w["y"])**2

def distToSegmentSquared(p, v, w):
  l2 = dist2(v, w)
  if l2 == 0:
    return dist2(p, v)
  t = ((p["x"] - v["x"]) * (w["x"] - v["x"]) + (p["y"] - v["y"]) * (w["y"] - v["y"])) / l2
  t = max(0, min(1, t))
  return dist2(p, {"x": v["x"] + t * (w["x"] - v["x"]), "y": v["y"] + t * (w["y"] - v["y"]) })

def distToSegment(p, v, w):
  return math.sqrt(distToSegmentSquared(p, v, w))

def checkNewPointNotRedundant(points, x, y):
  if len(points[0]) > 3:
    dist = (points[0] - x)**2 + (points[1] - y)**2
    min1 = np.argmin(dist)
    minDist = dist[min1]
    
    p = {"x": x, "y": y}
    for i in range(0,len(points[0])-1):
      v = {"x": points[0][i],   "y": points[1][i]   }
      w = {"x": points[0][i+1], "y": points[1][i+1] }
      d = distToSegment(p, v, w)
      if d < minDist:
        minDist = d
    
    if minDist <= 1:
      return 0
    else:
      return 1
  else:
    return 1

def appendPoint(x, y, points):
  curPoint = np.zeros((2, 1))
  curPoint[0] = x;
  curPoint[1] = y;
  points = np.append(points, curPoint, axis=1)
  return points

def recenterPointAlongOrthogonalTailAxis(x,y,theta,frame,thresh1):
  l = [[x,y]]
  
  thresh1lenX = len(thresh1[0]) - 1
  thresh1lenY = len(thresh1) - 1
  
  pixVal = 1
  k1 = 1
  while pixVal < 150 and k1 < 100:
    xNew = assignValueIfBetweenRange(int(x + k1 * (math.cos(theta)) ), 0, thresh1lenX)
    yNew = assignValueIfBetweenRange(int(y + k1 * (math.sin(theta)) ), 0, thresh1lenY)
    pixVal = thresh1[yNew][xNew]
    if pixVal < 150:
      k1 = k1 + 1
      cv2.circle(frame, (xNew, yNew), 1, (0,0,0),   -1)
      point = [xNew, yNew]
      l.append(point)

  pixVal = 1
  k2 = 1
  while pixVal < 150 and k2 < 100:
    xNew = assignValueIfBetweenRange(int(x - k2 * (math.cos(theta)) ), 0, thresh1lenX)
    yNew = assignValueIfBetweenRange(int(y - k2 * (math.sin(theta)) ), 0, thresh1lenY)
    pixVal = thresh1[yNew][xNew]
    if pixVal < 150:
      k2 = k2 + 1
      cv2.circle(frame, (xNew, yNew), 1, (0,0,0),   -1)
      point = [xNew, yNew]
      l = [point] + l
  
  n = len(l)
  xNew = assignValueIfBetweenRange(int((l[0][0] + l[n-1][0]) / 2), 0, thresh1lenX)
  yNew = assignValueIfBetweenRange(int((l[0][1] + l[n-1][1]) / 2), 0, thresh1lenY)
  
  return [xNew, yNew]

def findNextPoints(x,y,thresh1,frame,depth,points,lastTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,step,debugTracking):

  thresh1lenX = len(thresh1[0]) - 1
  thresh1lenY = len(thresh1) - 1
  
  maxThetaCorrected = -1

  if debugAdv:
    cv2.imshow('Frame', frame)
    cv2.waitKey(0)

  if (depth < 25): #15):
    maxTheta = 0
    ktotMax = 0
    k1Max = 0
    k2Max = 0
    
    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if abs(i*(math.pi/nbList)-lastTheta) < thetaDiffAccept]
      
    for thetaB in l:
      theta = thetaB % math.pi
      
      # Find furtherest point away from "old" "current" point still inside blob along the theta angle in one direction
      pixVal = 1
      k1 = 1
      while pixVal < 150 and k1 < 100:
        xNew = assignValueIfBetweenRange(int(x + k1 * (math.cos(theta))), 0, thresh1lenX)
        yNew = assignValueIfBetweenRange(int(y + k1 * (math.sin(theta))), 0, thresh1lenY)
        pixVal = thresh1[yNew][xNew]
        k1 = k1 + 1
      k1 = k1 - 2
      
      # Find furtherest point away from "old" "current" point still inside blob along the theta angle in the other direction
      pixVal = 1
      k2 = 1
      while pixVal < 150 and k2 < 100:
        xNew = assignValueIfBetweenRange(int(x - k2 * (math.cos(theta)) ), 0, thresh1lenX)
        yNew = assignValueIfBetweenRange(int(y - k2 * (math.sin(theta)) ), 0, thresh1lenY)
        pixVal = thresh1[yNew][xNew]
        k2 = k2 + 1
      k2 = k2 - 2
      ktot = k1 + k2
      
      # Calculate "score" for this theta angle based on the lenght of the segment fitted inside the blob
      ktot = 0
      for i in range(0,k1):
        ktot = ktot + math.exp(-i*expDecreaseFactor)
      for i in range(0,k2):
        ktot = ktot + math.exp(-i*expDecreaseFactor)
      
      # Keeps that theta angle as maximum if appropriate
      if (ktot > ktotMax):
        ktotMax = ktot
        k1Max = k1
        k2Max = k2
        maxTheta = theta
        
    # Calculates the two new possible points on both side of the local tail angle derivate
    if k1Max > step:
      x1 = assignValueIfBetweenRange(int(x + step  * (math.cos(maxTheta)) ), 0, thresh1lenX)
      y1 = assignValueIfBetweenRange(int(y + step  * (math.sin(maxTheta)) ), 0, thresh1lenY)
    else:
      x1 = assignValueIfBetweenRange(int(x + k1Max  * (math.cos(maxTheta)) ), 0, thresh1lenX)
      y1 = assignValueIfBetweenRange(int(y + k1Max  * (math.sin(maxTheta)) )  , 0, thresh1lenY)
    if k2Max > step:
      x2 = assignValueIfBetweenRange(int(x - step  * (math.cos(maxTheta)) ), 0, thresh1lenX)
      y2 = assignValueIfBetweenRange(int(y - step  * (math.sin(maxTheta)) ), 0, thresh1lenY)
    else:
      x2 = assignValueIfBetweenRange(int(x - k2Max  * (math.cos(maxTheta)) ), 0, thresh1lenX)
      y2 = assignValueIfBetweenRange(int(y - k2Max  * (math.sin(maxTheta)) ), 0, thresh1lenY)
    
    # Sanity check
    if debugTracking:
      if thresh1[y2][x2] > 150:
        print("oups white zone forw",k2Max)
      if thresh1[y1][x1] > 150:
        print("oups white zone back",k1Max)
        
    if (depth == 0):
      predictedTetha = maxTheta
    else:
      predictedTetha = maxTheta + ((maxTheta - lastTheta + 2*math.pi) % (2*math.pi))
    [x1, y1] = recenterPointAlongOrthogonalTailAxis(x1, y1, predictedTetha+(math.pi)/2, frame, thresh1)
    [x2, y2] = recenterPointAlongOrthogonalTailAxis(x2, y2, predictedTetha+(math.pi)/2, frame, thresh1)
    
    x1 = assignValueIfBetweenRange(x1, 0, thresh1lenX)
    y1 = assignValueIfBetweenRange(y1, 0, thresh1lenY)
    x2 = assignValueIfBetweenRange(x2, 0, thresh1lenX)
    y2 = assignValueIfBetweenRange(y2, 0, thresh1lenY)
    
    cv2.circle(frame, (x, y), 1, (255,0,255),   -1)
    
    # Calculates distance between new and old local tail angle
    diffAngle1 = distBetweenThetas(lastTheta, maxTheta )
    diffAngle2 = distBetweenThetas(lastTheta, (maxTheta + math.pi) )
    
    # Calculates distance between new and old point
    distSubsquentPoints1 = (x1 - x)**2 + (y1 - y)**2
    distSubsquentPoints2 = (x2 - x)**2 + (y2 - y)**2
    
    # If the distance between new and old point larger than 0, appends points and launch search for new point
    if (diffAngle1 > diffAngle2):
      if debugTracking:
        if thresh1[y2][x2] > 150:
          print("oups white zone")
        if debugAdv:
          print("diffAngle2: ",diffAngle2," ; k2Max: ",k2Max," distSubsquentPoints2:",distSubsquentPoints2)
      cv2.circle(frame, (x2, y2), 1, (0,255,0),   -1)
      check = checkNewPointNotRedundant(points, x2, y2)
      if check:
        points = appendPoint(x2, y2, points)
        if distSubsquentPoints2 > 0:
          maxThetaCorrected = maxTheta + math.pi
          newTheta = calculateAngle(x,y,x2,y2)
          (points,nop) = findNextPoints(x2,y2,thresh1,frame,depth+1,points,newTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,step,debugTracking)
    else:
      if debugTracking:
        if thresh1[y1][x1] > 150:
          print("oups white zone")
        if debugAdv:
          print("diffAngle1: ",diffAngle1," ; k1Max: ",k1Max," distSubsquentPoints1:",distSubsquentPoints1)
      cv2.circle(frame, (x1, y1), 1, (0,255,0),   -1)
      check = checkNewPointNotRedundant(points, x1, y1)
      if check:
        points = appendPoint(x1, y1, points)
        if distSubsquentPoints1 > 0:
          maxThetaCorrected = maxTheta
          newTheta = calculateAngle(x,y,x1,y1)
          (points,nop) = findNextPoints(x1,y1,thresh1,frame,depth+1,points,newTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,step,debugTracking)
        
  return (points,maxThetaCorrected)
  
def smoothTail(points, nbTailPoints):
  
  y = points[0]
  x = linspace(0, 1, len(y))
  
  if len(x) > 3:    
  
    s = UnivariateSpline(x, y, s=10)
    xs = linspace(0, 1, nbTailPoints)
    newX = s(xs)

    y = points[1]
    x = linspace(0, 1, len(y))
    s = UnivariateSpline(x, y, s=10)
    xs = linspace(0, 1, nbTailPoints)
    newY = s(xs)
    
  else:
  
    newX = x
    newY = y
  
  return [newX, newY]

  # x,y,thresh1,frame,0,points,lastFirstTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,ste,debugTracking
  
def tailTrackingBlobDescent(headPosition,nbTailPoints,i,x,y,thresh1,frame,lastFirstTheta,debugAdv,thetaDiffAccept,hyperparameters):

  step = hyperparameters["step"]
  nbList = hyperparameters["nbList"]
  expDecreaseFactor = hyperparameters["expDecreaseFactor"]
  firstFrame = hyperparameters["firstFrame"]
  lastFrame = hyperparameters["lastFrame"]
  minArea = hyperparameters["minArea"]
  maxArea = hyperparameters["maxArea"]
  headSize = hyperparameters["headSize"]
  debugTracking = hyperparameters["debugTracking"]
  headEmbeded = hyperparameters["headEmbeded"]
  
  output = np.zeros((1, nbTailPoints, 2))
  
  possiblePoints = []
  for ste in step:
    points = np.zeros((2, 0))
    (points, lastFirstTheta2) = findNextPoints(x,y,thresh1,frame,0,points,lastFirstTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,ste,debugTracking)
    possiblePoints.append(points)
  
  nbPossiblePoints = len(possiblePoints)
  minDist = 1000
  pBest = 0
  if nbPossiblePoints == 1:
    points = possiblePoints[0]
  else:
    for p1 in range(0, nbPossiblePoints):
      for p2 in range(0, nbPossiblePoints):
        if p2 > p1:
          points1 = possiblePoints[p1]
          n1 = len(points1[0]) - 1
          x1 = int(points1[0][n1])
          y1 = int(points1[1][n1])
          
          points2 = possiblePoints[p2]
          n2 = len(points2[0]) - 1
          x2 = int(points2[0][n2])
          y2 = int(points2[1][n2])
          
          dist = math.sqrt((x1-x2)**2 + (y1-y2)**2)
          if dist < minDist:
            minDist = dist
            pBest = p1
    points = possiblePoints[pBest]
  
  points = np.insert(points, 0, headPosition, axis=1) 
  points = smoothTail(points, nbTailPoints)

  # for idx, x in enumerate(points[0]):
    # output[i-firstFrame][idx][0] = x
    # output[i-firstFrame][idx][1] = points[1][idx]
    
  for idx, x in enumerate(points[0]):
    output[0][idx][0] = x
    output[0][idx][1] = points[1][idx]

  return output



from zebrazoom.code.getHyperparameters import getHyperparametersSimple
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import json
import numpy as np
import sys
import os
import cv2
from pathlib import Path

try:
  from PyQt6.QtCore import Qt, QSize, QTimer
  from PyQt6.QtWidgets import QLabel, QSlider, QVBoxLayout
except ImportError:
  from PyQt5.QtCore import Qt, QSize, QTimer
  from PyQt5.QtWidgets import QLabel, QSlider, QVBoxLayout

import zebrazoom.code.util as util


def readValidationVideo(videoPath, folderName, configFilePath, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation=''):
  s1  = "ZZoutput"
  s2  = folderName
  s3b = "results_"
  s4  = folderName
  s5  = ".avi"
  s5b = ".txt"

  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  initialPath  = Path(cur_dir_path)
  initialPath  = initialPath.parent
  initialPath  = os.path.join(initialPath, s1)
  if len(ZZoutputLocation):
    initialPath = ZZoutputLocation

  with open(os.path.join(initialPath, os.path.join(s2, 'configUsed.json'))) as f:
    configTemp = json.load(f)
  hyperparameters = getHyperparametersSimple(configTemp)

  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] and os.path.exists(os.path.join(initialPath, os.path.join(s1, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi')))):
    # The "exist" check above is only to insure compatibility with videos tracked prior to this update
    videoPath = os.path.join(initialPath, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'))
  else:
    videoPath = os.path.join(initialPath, os.path.join(s2, s4 + s5))

  resultsPath = os.path.join(initialPath, os.path.join(s2, s3b + s4 + s5b))

  if not(os.path.exists(videoPath)):
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if '.avi' in fileName:
        resultFile = fileName
    videoPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  if not(os.path.exists(resultsPath)):
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if 'results_' in fileName:
        resultFile = fileName
    resultsPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  cap = zzVideoReading.VideoCapture(videoPath)

  nx    = int(cap.get(3))
  ny    = int(cap.get(4))
  max_l = int(cap.get(7))

  with open(resultsPath) as f:
    supstruct = json.load(f)

  if not("firstFrame" in supstruct):
    supstruct["firstFrame"] = 1
    print("supstruct['firstFrame'] not found")

  infoWells = []

  HeadX = np.zeros(max_l + supstruct["firstFrame"])
  HeadY = np.zeros(max_l + supstruct["firstFrame"])

  if ((numWell != -1) and (zoom)):
    lastEnd = 0
    lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadX"][0]
    lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadY"][0]
    for k in range(0,len(supstruct["wellPoissMouv"][numWell][numAnimal])):
      beg = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutStart"]
      end = supstruct["wellPoissMouv"][numWell][numAnimal][k]["BoutEnd"]
      for l in range(lastEnd, beg):
        HeadX[l] = lastXpos
        HeadY[l] = lastYpos
      for l in range(beg, end):
        HeadX[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][l-beg]
        HeadY[l]  = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][l-beg]
      lastEnd = end
      lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadX"][end-1-beg]
      lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][k]["HeadY"][end-1-beg]

    for l in range(lastEnd, max_l + supstruct["firstFrame"]):
      HeadX[l] = lastXpos
      HeadY[l] = lastYpos

  # /* Getting the info about well positions */
  analyzeAllWellsAtTheSameTime = 0
  if (analyzeAllWellsAtTheSameTime == 0):
    for i in range(0, len(supstruct["wellPositions"])):
      x = 0
      y = 0
      lengthX = 0
      lengthY = 0
      rectangleWellArea = 1
      if (rectangleWellArea == 0): # circular wells
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        r = supstruct["wellPositions"][i]["diameter"]
        lengthX = 300 # wellOutputVideoDiameter;
        lengthY = 300 # wellOutputVideoDiameter;
      else:
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        lengthX = supstruct["wellPositions"][i]["lengthX"]
        lengthY = supstruct["wellPositions"][i]["lengthY"]
      if (x < 0):
        x = 0
      if (y < 0):
        y = 0
      infoWells.append([x, y, lengthX, lengthY])
  else:
    infoWells.append([0, 0, nx, ny])

  frameToPosToPlot = None
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"]:
    frameToPosToPlot = {}
    for frameNumber in range(l, max_l + 400):
      frameToPosToPlot[frameNumber] = []
    for numWell in range(0, len(supstruct["wellPoissMouv"])):
      for numAnimal in range(0, len(supstruct["wellPoissMouv"][numWell])):
        for numBout in range(0, len(supstruct["wellPoissMouv"][numWell][numAnimal])):
          boutStart = supstruct["wellPoissMouv"][numWell][numAnimal][numAnimal]["BoutStart"]
          for i in range(0, len(supstruct["wellPoissMouv"][numWell][numAnimal][numAnimal]["HeadX"])):
            if boutStart + i in frameToPosToPlot:
              if (type(framesToShow) != np.ndarray) or (framesToShow[boutStart + i][numWell]):
                xPos = int(supstruct["wellPositions"][numWell]["topLeftX"] + supstruct["wellPoissMouv"][numWell][numAnimal][numAnimal]["HeadX"][i])
                yPos = int(supstruct["wellPositions"][numWell]["topLeftY"] + supstruct["wellPoissMouv"][numWell][numAnimal][numAnimal]["HeadY"][i])
                frameToPosToPlot[boutStart + i].append([xPos, yPos])

  x = 0
  y = 0
  lengthX = 0
  lengthY = 0
  if (numWell != -1):
    x = infoWells[numWell][0]
    y = infoWells[numWell][1]
    lengthX = infoWells[numWell][2]
    lengthY = infoWells[numWell][3]
  else:
    lengthX = nx
    lengthY = ny

  l = start - supstruct["firstFrame"] + 1 if start > 0 else 0

  xOriginal = x
  yOriginal = y

  def getFrame():
    nonlocal x
    nonlocal y
    nonlocal lengthX
    nonlocal lengthY

    l = frameSlider.value()
    if timer.isActive() and (l == frameSlider.maximum() or stopTimer):
      timer.stop()

    cap.set(1, l)
    ret, img = cap.read()

    if frameToPosToPlot is not None:
      if l in frameToPosToPlot:
        for pos in frameToPosToPlot[l]:
          cv2.circle(img, (pos[0], pos[1]), hyperparameters["trackingPointSizeDisplay"], (0, 255, 0), -1)

    if numWell != -1 and zoom:
      length = 250
      xmin = int(HeadX[l + supstruct["firstFrame"] - 1] - length/2)
      xmax = int(HeadX[l + supstruct["firstFrame"] - 1] + length/2)
      ymin = int(HeadY[l + supstruct["firstFrame"] - 1] - length/2)
      ymax = int(HeadY[l + supstruct["firstFrame"] - 1] + length/2)

      x = max(xmin + xOriginal, 0)
      y = max(ymin + yOriginal, 0)
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      if y + lengthY >= len(img):
        lengthY = len(img) - y - 1
      if x + lengthX >= len(img[0]):
        lengthX = len(img[0]) - x - 1

    if (numWell != -1):
      img = img[y:y+lengthY, x:x+lengthX]

    if lengthX > 100 and lengthY > 100:
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img,str(l + supstruct["firstFrame"] - 1),(int(lengthX-110), int(lengthY-30)),font,1,(0,255,0))
    else:
      blank_image = np.zeros((len(img)+30, len(img[0]), 3), np.uint8)
      blank_image[0:len(img), 0:len(img[0])] = img
      img = blank_image
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img, str(l + supstruct["firstFrame"] - 1), (int(0), int(lengthY+25)), font, 1, (0,255,0))

    return img

  layout = QVBoxLayout()

  video = QLabel()
  layout.addWidget(video, stretch=1)

  frameSlider = QSlider(Qt.Orientation.Horizontal)
  frameSlider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
  frameSlider.setPageStep(50)
  frameSlider.setRange(0, max_l - 1)
  frameSlider.setValue(l)
  frameSlider.valueChanged.connect(lambda: util.setPixmapFromCv(getFrame(), video))
  layout.addWidget(frameSlider)
  shortcutsLabel = QLabel("Left Arrow, Right Arrow, Page Up, Page Down, Home and End keys can be used to navigate through the video.")
  shortcutsLabel.setWordWrap(True)
  layout.addWidget(shortcutsLabel)

  stopTimer = True
  timer = QTimer()
  timer.setInterval(1)

  def nextFrame():
    nonlocal stopTimer
    stopTimer = False
    frameSlider.setValue(frameSlider.value() + 1)
    stopTimer = True
  timer.timeout.connect(nextFrame)

  startFrame = getFrame()
  timer.start()
  util.showDialog(layout, title="Video", labelInfo=(startFrame, video))
  timer.stop()
  cap.release()



import sys
try:
  from PyQt6.QtCore import Qt, QEventLoop
  from PyQt6.QtGui import QFont
  from PyQt6.QtWidgets import QWidget, QApplication, QLabel, QPushButton, QLineEdit, QGridLayout, QRadioButton
except ImportError:
  from PyQt5.QtCore import Qt, QEventLoop
  from PyQt5.QtGui import QFont
  from PyQt5.QtWidgets import QWidget, QApplication, QLabel, QPushButton, QLineEdit, QGridLayout, QRadioButton

from pathlib import Path
import pandas as pd
import subprocess
import json
import sys
import os

from zebrazoom.kinematicParametersAnalysis import kinematicParametersAnalysis

def getVideoFPSandPixelSizeThroughGUI(initialFPS, initialPixelSize, videoName):
  
  loop = QEventLoop()
  window = QWidget()
  window.setWindowTitle('Choose video fps and pixel size')
  window.setGeometry(100, 100, 280, 80)
  window.move(60, 15)

  layout = QGridLayout()
  videoFPS = QLineEdit(str(initialFPS))
  videoPixelSize = QLineEdit(str(initialPixelSize))
  setSameForAll = QRadioButton('set all videos to the same video fps and pixel size')
  okbutton = QPushButton('Ok')
  okbutton.clicked.connect(lambda: loop.exit())

  layout.addWidget(QLabel('<center><h3>' + videoName + '</h3></center>'), 0, 0, 1, 3)
  layout.addWidget(QLabel('videoFPS:'), 1, 0)
  layout.addWidget(videoFPS, 1, 1)
  layout.addWidget(QLabel('in Hz (1/second)'), 1, 2)
  layout.addWidget(QLabel('videoPixelSize:'), 2, 0)
  layout.addWidget(videoPixelSize, 2, 1)
  layout.addWidget(QLabel('in mm/pixel'), 2, 2)
  layout.addWidget(setSameForAll, 3, 0, 1, 3)
  layout.addWidget(okbutton, 4, 0, 1, 3)
  window.setLayout(layout)
  window.show()
  loop.exec()
  window.close()
  return [float(videoFPS.text()), float(videoPixelSize.text()), int(setSameForAll.isChecked())]

def checkConsistencyOfParameters(listOfVideosToCheckConsistencyOn):
  
  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  cur_dir_path = Path(cur_dir_path)
  zebrazoom_path = cur_dir_path.parent
  app = QApplication.instance()
  ZZoutputFolder = os.path.join(zebrazoom_path, 'ZZoutput') if app is None else app.ZZoutputLocation
  
  data = []
  
  videoFPS_forAllVideos       = -1
  videoPixelSize_forAllVideos = -1
  setSameForAll = False
  for videoName in listOfVideosToCheckConsistencyOn:
    with open(os.path.join(ZZoutputFolder, videoName, 'configUsed.json')) as f:
      configFile = json.load(f)
    nbWells        = configFile["nbWells"]
    # Checking / getting information about videoFPS and videoPixelSize
    if not(setSameForAll):
      videoFPS       = configFile["videoFPS"] if "videoFPS" in configFile else videoFPS_forAllVideos
      videoPixelSize = configFile["videoPixelSize"] if "videoPixelSize" in configFile else videoPixelSize_forAllVideos
      [videoFPS, videoPixelSize, setSameForAll] = getVideoFPSandPixelSizeThroughGUI(videoFPS, videoPixelSize, videoName)
      if setSameForAll:
        videoFPS_forAllVideos       = videoFPS
        videoPixelSize_forAllVideos = videoPixelSize
    else:
      videoFPS       = videoFPS_forAllVideos
      videoPixelSize = videoPixelSize_forAllVideos
    # Setting the videoFPS and videoPixelSize values inside the excel file
    vidTab  = ["defaultZZoutputFolder" if app is None else ZZoutputFolder, videoName, videoFPS, videoPixelSize, str([1 for i in range(nbWells)]), str(['Your data' for i in range(nbWells)]), str([1 for i in range(nbWells)])]
    data.append(vidTab)
  
  nbWellsStandardValues = 100
  data.append(["defaultZZoutputFolder", 'standardValueFreelySwimZebrafishLarvae', 1, 1, str([1 for i in range(nbWellsStandardValues)]), str(['StandardValues' for i in range(nbWellsStandardValues)]), str([1 for i in range(nbWellsStandardValues)])])
  excelFileDataFrame = pd.DataFrame(data=data, columns=['path', 'trial_id', 'fq', 'pixelsize', 'condition', 'genotype', 'include'])
  excelFileDataFrame.to_excel(os.path.join(os.path.join(os.path.join(zebrazoom_path, 'dataAnalysis'), 'experimentOrganizationExcel'), 'tempExcelFileForParametersConsistencyCheck.xls'))
  class sysSimulation:
    argv = []
  sysSimul = sysSimulation()
  sysSimul.argv = ['', '', '', os.path.join(os.path.join(os.path.join(zebrazoom_path, 'dataAnalysis'), 'experimentOrganizationExcel'), 'tempExcelFileForParametersConsistencyCheck.xls'), 4, -1, 1, -1, 0]
  kinematicParametersAnalysis(sysSimul, 0, True)
  dir_path = os.path.join(os.path.join(os.path.join(zebrazoom_path, 'dataAnalysis'), 'resultsKinematic'), 'tempExcelFileForParametersConsistencyCheck')
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])



import cv2
import re
import os
import json
import shutil
import sys
import subprocess
from matplotlib.figure import Figure
import math
import scipy.io as sio
from pathlib import Path
from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()

from zebrazoom.mainZZ import mainZZ
from zebrazoom.getTailExtremityFirstFrame import getTailExtremityFirstFrame
import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow
import zebrazoom.code.util as util

try:
  from PyQt6.QtCore import Qt
  from PyQt6.QtWidgets import QCheckBox, QFileDialog, QVBoxLayout
except ImportError:
  from PyQt5.QtCore import Qt
  from PyQt5.QtWidgets import QCheckBox, QFileDialog, QVBoxLayout


LARGE_FONT= ("Verdana", 12)

def chooseVideoToAnalyze(self, justExtractParams, noValidationVideo, chooseFrames, testMode):
    self.videoName, _ = QFileDialog.getOpenFileName(self.window, 'Select file', os.path.expanduser("~"))
    if not self.videoName:
      return
    self.folderName = ''
    self.headEmbedded = 0
    self.sbatchMode = 0

    self.justExtractParams = int(justExtractParams)
    self.noValidationVideo = int(noValidationVideo)
    self.testMode         = testMode
    self.findMultipleROIs = 0

    if chooseFrames:
      def beginningAndEndChosen():
        if "firstFrame" in self.configFile:
          self.firstFrame = self.configFile["firstFrame"]
        if "lastFrame" in self.configFile:
          self.lastFrame = self.configFile["lastFrame"]
        self.configFile.clear()
        self.backgroundExtractionForceUseAllVideoFrames = int(backgroundExtractionForceUseAllVideoFramesCheckbox.isChecked())
        self.show_frame("ConfigFilePromp")
      layout = QVBoxLayout()
      backgroundExtractionForceUseAllVideoFramesCheckbox = QCheckBox("Use all frames to calculate background")
      backgroundExtractionForceUseAllVideoFramesCheckbox.setChecked(True)
      layout.addWidget(backgroundExtractionForceUseAllVideoFramesCheckbox, alignment=Qt.AlignmentFlag.AlignCenter)
      util.chooseBeginningPage(self, self.videoName, "Choose where the analysis of your video should start.", "Ok, I want the tracking to start at this frame!",
                               lambda: util.chooseEndPage(self, self.videoName, "Choose where the analysis of your video should end.", "Ok, I want the tracking to end at this frame!", beginningAndEndChosen),
                               additionalLayout=layout)
    else:
      self.show_frame("ConfigFilePromp")

def chooseFolderToAnalyze(self, justExtractParams, noValidationVideo, sbatchMode):
    self.folderName =  QFileDialog.getExistingDirectory(self.window, 'Select folder', os.path.expanduser("~"))
    if not self.folderName:
      return
    self.headEmbedded = 0
    self.justExtractParams = int(justExtractParams)
    self.noValidationVideo = int(noValidationVideo)
    self.sbatchMode        = int(sbatchMode)
    self.testMode = False
    self.findMultipleROIs = 0
    self.show_frame("ConfigFilePromp")

def chooseFolderForTailExtremityHE(self):
    self.folderName =  QFileDialog.getExistingDirectory(self.window, 'Select folder', os.path.expanduser("~"))
    if not self.folderName:
      return
    self.sbatchMode = 0
    self.headEmbedded = 1
    self.justExtractParams = 0
    self.noValidationVideo = 0
    self.testMode = False
    self.findMultipleROIs = 0
    self.show_frame("ConfigFilePromp")

def chooseFolderForMultipleROIs(self):
    self.folderName =  QFileDialog.getExistingDirectory(self.window, 'Select folder', os.path.expanduser("~"))
    if not self.folderName:
      return
    self.sbatchMode = 0
    self.headEmbedded = 0
    self.justExtractParams = 0
    self.noValidationVideo = 0
    self.testMode = False
    self.findMultipleROIs = 1
    self.show_frame("ConfigFilePromp")

def chooseConfigFile(self):

  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  path = Path(cur_dir_path)
  path = path.parent.parent
  path = os.path.join(path, 'configuration')

  self.configFileName, _ = QFileDialog.getOpenFileName(self.window, 'Select file', path, "JSON (*.json)")
  if not self.configFileName:
    return
  if len(self.folderName) or globalVariables["mac"] or globalVariables["lin"]:
    self.show_frame("Patience")
  else:
    self.launchZebraZoom()

def findAllFilesRecursivelyInDirectories(folderName):

  allVideos = []

  for name in os.listdir(folderName):
    if os.path.isdir(os.path.join(folderName, name)):
      allVideos2 = findAllFilesRecursivelyInDirectories(os.path.join(folderName, name))
      allVideos = allVideos + allVideos2
    else:
      if len(name) > 3:
        ext = os.path.splitext(name)[1]
        if len(ext):
          if (ext in ['.264', '.3g2', '.3gp', '.3gp2', '.3gpp', '.3gpp2', '.3mm', '.3p2', '.60d', '.787', '.89', '.aaf', '.aec', '.aep', '.aepx', '.aet', '.aetx', '.ajp', '.ale', '.am', '.amc', '.amv', '.amx', '.anim', '.aqt', '.arcut', '.arf', '.asf', '.asx', '.avb', '.avc', '.avd', '.avi', '.avp', '.avs', '.avs', '.avv', '.axm', '.bdm', '.bdmv', '.bdt2', '.bdt3', '.bik', '.bix', '.bmk', '.bnp', '.box', '.bs4', '.bsf', '.bvr', '.byu', '.camproj', '.camrec', '.camv', '.ced', '.cel', '.cine', '.cip', '.clpi', '.cmmp', '.cmmtpl', '.cmproj', '.cmrec', '.cpi', '.cst', '.cvc', '.cx3', '.d2v', '.d3v', '.dat', '.dav', '.dce', '.dck', '.dcr', '.dcr', '.ddat', '.dif', '.dir', '.divx', '.dlx', '.dmb', '.dmsd', '.dmsd3d', '.dmsm', '.dmsm3d', '.dmss', '.dmx', '.dnc', '.dpa', '.dpg', '.dream', '.dsy', '.dv', '.dv-avi', '.dv4', '.dvdmedia', '.dvr', '.dvr-ms', '.dvx', '.dxr', '.dzm', '.dzp', '.dzt', '.edl', '.evo', '.eye', '.ezt', '.f4p', '.f4v', '.fbr', '.fbr', '.fbz', '.fcp', '.fcproject', '.ffd', '.flc', '.flh', '.fli', '.flv', '.flx', '.gfp', '.gl', '.gom', '.grasp', '.gts', '.gvi', '.gvp', '.h264', '.hdmov', '.hkm', '.ifo', '.imovieproj', '.imovieproject', '.ircp', '.irf', '.ism', '.ismc', '.ismv', '.iva', '.ivf', '.ivr', '.ivs', '.izz', '.izzy', '.jss', '.jts', '.jtv', '.k3g', '.kmv', '.ktn', '.lrec', '.lsf', '.lsx', '.m15', '.m1pg', '.m1v', '.m21', '.m21', '.m2a', '.m2p', '.m2t', '.m2ts', '.m2v', '.m4e', '.m4u', '.m4v', '.m75', '.mani', '.meta', '.mgv', '.mj2', '.mjp', '.mjpg', '.mk3d', '.mkv', '.mmv', '.mnv', '.mob', '.mod', '.modd', '.moff', '.moi', '.moov', '.mov', '.movie', '.mp21', '.mp21', '.mp2v', '.mp4', '.mp4v', '.mpe', '.mpeg', '.mpeg1', '.mpeg4', '.mpf', '.mpg', '.mpg2', '.mpgindex', '.mpl', '.mpl', '.mpls', '.mpsub', '.mpv', '.mpv2', '.mqv', '.msdvd', '.mse', '.msh', '.mswmm', '.mts', '.mtv', '.mvb', '.mvc', '.mvd', '.mve', '.mvex', '.mvp', '.mvp', '.mvy', '.mxf', '.mxv', '.mys', '.ncor', '.nsv', '.nut', '.nuv', '.nvc', '.ogm', '.ogv', '.ogx', '.osp', '.otrkey', '.pac', '.par', '.pds', '.pgi', '.photoshow', '.piv', '.pjs', '.playlist', '.plproj', '.pmf', '.pmv', '.pns', '.ppj', '.prel', '.pro', '.prproj', '.prtl', '.psb', '.psh', '.pssd', '.pva', '.pvr', '.pxv', '.qt', '.qtch', '.qtindex', '.qtl', '.qtm', '.qtz', '.r3d', '.rcd', '.rcproject', '.rdb', '.rec', '.rm', '.rmd', '.rmd', '.rmp', '.rms', '.rmv', '.rmvb', '.roq', '.rp', '.rsx', '.rts', '.rts', '.rum', '.rv', '.rvid', '.rvl', '.sbk', '.sbt', '.scc', '.scm', '.scm', '.scn', '.screenflow', '.sec', '.sedprj', '.seq', '.sfd', '.sfvidcap', '.siv', '.smi', '.smi', '.smil', '.smk', '.sml', '.smv', '.spl', '.sqz', '.srt', '.ssf', '.ssm', '.stl', '.str', '.stx', '.svi', '.swf', '.swi', '.swt', '.tda3mt', '.tdx', '.thp', '.tivo', '.tix', '.tod', '.tp', '.tp0', '.tpd', '.tpr', '.trp', '.ts', '.tsp', '.ttxt', '.tvs', '.usf', '.usm', '.vc1', '.vcpf', '.vcr', '.vcv', '.vdo', '.vdr', '.vdx', '.veg','.vem', '.vep', '.vf', '.vft', '.vfw', '.vfz', '.vgz', '.vid', '.video', '.viewlet', '.viv', '.vivo', '.vlab', '.vob', '.vp3', '.vp6', '.vp7', '.vpj', '.vro', '.vs4', '.vse', '.vsp', '.w32', '.wcp', '.webm', '.wlmp', '.wm', '.wmd', '.wmmp', '.wmv', '.wmx', '.wot', '.wp3', '.wpl', '.wtv', '.wve', '.wvx', '.xej', '.xel', '.xesc', '.xfl', '.xlmv', '.xmv', '.xvid', '.y4m', '.yog', '.yuv', '.zeg', '.zm1', '.zm2', '.zm3', '.zmv']):
            allVideos.append(os.path.join(folderName, name))

  return allVideos


def launchZebraZoom(self):
  if self.testMode:
    with open(self.configFileName) as f:
      self.configFile = json.load(f)
    self.videoToCreateConfigFileFor = self.videoName
    self.testConfig(addToHistory=False)
    self.headEmbedded      = 0
    self.justExtractParams = 0
    self.noValidationVideo = 0
    self.testMode         = False
    self.findMultipleROIs  = 0
    self.configFileName = None
    self.videoName = None
    return

  last = 0
  allVideos = []

  if self.sbatchMode:
    commandsFile = open("commands.txt", "w", newline='\n')
    nbVideosToLaunch = 0

  if len(self.folderName):

    allVideos = findAllFilesRecursivelyInDirectories(self.folderName)

  else:
    allVideos = [self.videoName]

  print("allVideos:", allVideos)

  for idx, text in enumerate(allVideos):

    path        = os.path.split(text)[0]
    nameWithExt = os.path.split(text)[1]
    name        = os.path.splitext(nameWithExt)[0]
    videoExt    = os.path.splitext(nameWithExt)[1][1:]

    if self.headEmbedded == 0:
      if len(allVideos) == 1:
        tabParams = ["mainZZ", path, name, videoExt, self.configFileName, "freqAlgoPosFollow", 100, "popUpAlgoFollow", 1, "outputFolder", self.ZZoutputLocation]
      else:
        tabParams = ["mainZZ", path, name, videoExt, self.configFileName, "freqAlgoPosFollow", 100, "outputFolder", self.ZZoutputLocation]
      if hasattr(self, "backgroundExtractionForceUseAllVideoFrames"):
        tabParams.extend(["backgroundExtractionForceUseAllVideoFrames", self.backgroundExtractionForceUseAllVideoFrames])
      if hasattr(self, "firstFrame"):
        tabParams.extend(["firstFrame", self.firstFrame])
      if hasattr(self, "lastFrame"):
        tabParams.extend(["lastFrame", self.lastFrame])
      if self.justExtractParams == 1:
        tabParams = tabParams + ["reloadWellPositions", 1, "reloadBackground", 1, "debugPauseBetweenTrackAndParamExtract", "justExtractParamFromPreviousTrackData"]
      if self.noValidationVideo == 1:
          tabParams = tabParams + ["createValidationVideo", 0]
      if self.findMultipleROIs == 1:
        tabParams = tabParams + ["exitAfterWellsDetection", 1, "saveWellPositionsToBeReloadedNoMatterWhat", 1]
      try:
        if self.sbatchMode:
          commandsFile.write('python -m zebrazoom ' + ' '.join(tabParams[1:4]).replace('\\', '/').replace('//lexport/iss01.', '/network/lustre/iss01/') + ' configFile.json\n')
          nbVideosToLaunch = nbVideosToLaunch + 1
        else:
          mainZZ(path, name, videoExt, self.configFileName, tabParams)
      except ValueError:
        print("moving on to the next video for ROIs identification")
      except NameError:
        self.show_frame("Error")
        return
    else:
      tabParams = ["outputFolder", self.ZZoutputLocation]
      if hasattr(self, "backgroundExtractionForceUseAllVideoFrames"):
        tabParams.extend(["backgroundExtractionForceUseAllVideoFrames", self.backgroundExtractionForceUseAllVideoFrames])
      if hasattr(self, "firstFrame"):
        tabParams.extend(["firstFrame", self.firstFrame])
      if hasattr(self, "lastFrame"):
        tabParams.extend(["lastFrame", self.lastFrame])
      getTailExtremityFirstFrame(path, name, videoExt, self.configFileName, tabParams)

  self.headEmbedded      = 0
  self.justExtractParams = 0
  self.noValidationVideo = 0
  self.testMode         = False
  self.findMultipleROIs  = 0
  if hasattr(self, "backgroundExtractionForceUseAllVideoFrames"):
    del self.backgroundExtractionForceUseAllVideoFrames
  if hasattr(self, "firstFrame"):
    del self.firstFrame
  if hasattr(self, "lastFrame"):
    del self.lastFrame

  if self.sbatchMode:

    commandsFile.close()

    with open(self.configFileName) as f:
      jsonFile = json.load(f)
    nbWells = jsonFile["nbWells"]
    if nbWells > 24:
      nbWells = 24

    launchFile = open("launchZZ.sh", "w", newline='\n')
    linesToWrite = ['#!/bin/sh',
                    '#SBATCH --ntasks=1',
                    '#SBATCH --cpus-per-task='+str(nbWells),
                    '#SBATCH --array=1-'+str(nbVideosToLaunch),
                    '#SBATCH --mem=16G',
                    '#SBATCH --time=23:00:00',
                    '#SBATCH --partition=normal',
                    '#SBATCH --job-name="ZebraZoom-protocole"',
                    '',
                    'module load python/3.8',
                    'source activate zebrazoom',
                    '',
                    'date',
                    '',
                    'export CMD_FILE_PATH=./commands.txt',
                    '',
                    'export CMD=$(sed -n ${SLURM_ARRAY_TASK_ID}p ${CMD_FILE_PATH})',
                    '',
                    'echo $CMD',
                    'eval $CMD',
                    '',
                    'date']
    linesToWrite = [line + '\n' for line in linesToWrite]
    launchFile.writelines(linesToWrite)
    launchFile.close()

    shutil.copy(self.configFileName, 'configFile.json')

    self.show_frame("ZZoutroSbatch")

  else:

    self.show_frame("ZZoutro")


def openConfigurationFileFolder(self, homeDirectory):
  dir_path = os.path.join(homeDirectory,'configuration')
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])


def openZZOutputFolder(self, homeDirectory):
  if len(self.ZZoutputLocation):
    dir_path = self.ZZoutputLocation
  else:
    dir_path = os.path.join(homeDirectory,'ZZoutput')
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])



import zebrazoom
import pickle

# Creating the dataframe on which the clustering will be applied

dataframeOptions = {
  'pathToExcelFile'                   : './experimentOrganizationExcel/',
  'fileExtension'                     : '.xls',
  'resFolder'                         : 'data',
  'nameOfFile'                        : 'example',
  'smoothingFactorDynaParam'          : 0,   # 0.001
  'nbFramesTakenIntoAccount'          : -1, #28,
  'numberOfBendsIncludedForMaxDetect' : -1,
  'minNbBendForBoutDetect'            : 3,
  'defaultZZoutputFolderPath'         : '../ZZoutput/',
  'computeTailAngleParamForCluster'   : True,
  'computeMassCenterParamForCluster'  : True
}

[conditions, genotypes, nbFramesTakenIntoAccount, globParam] = zebrazoom.createDataFrame(dataframeOptions)


# Applying the clustering on this dataframe

clusteringOptions = {
  'analyzeAllWellsAtTheSameTime' : 0, # put this to 1 for head-embedded videos, and to 0 for multi-well videos
  'pathToVideos' : '../ZZoutput/',
  'nbCluster' : 3,
  'modelUsedForClustering' : 'KMeans', # put either 'KMeans' or 'GaussianMixture' here
  #'nbPcaComponents' : 30,
  'nbFramesTakenIntoAccount' : nbFramesTakenIntoAccount,
  'scaleGraphs' : True,
  'showFigures' : False,
  'useFreqAmpAsym' : False,
  'useAngles' : False,
  'useAnglesSpeedHeadingDisp' : False,
  'useAnglesSpeedHeading' : True,
  'useAnglesSpeed' : False,
  'useAnglesHeading' : False,
  'useAnglesHeadingDisp' : False,
  'useFreqAmpAsymSpeedHeadingDisp' : False,
  'videoSaveFirstTenBouts' : False,
  'globalParametersCalculations' : True,
  'nbVideosToSave' : 10,
  'resFolder' : 'data',
  'nameOfFile' : 'example'
}


# Applies the clustering for the first time
[allBouts, classifier] = zebrazoom.applyClustering(clusteringOptions, 0, 'resultsClustering')


# Saves the classifier
outfile = open('classifiers/classifier_' + clusteringOptions['nameOfFile'] + '.txt','wb')
pickle.dump([classifier, nbFramesTakenIntoAccount],outfile)
outfile.close()



from sklearn import metrics
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from modAL.models import ActiveLearner
from zebrazoom.dataAnalysis.dataanalysis.visualizeClusters import visualizeClusters
import matplotlib.pyplot as plt
from sklearn import cluster
from tkinter import simpledialog
import tkinter as tk
import numpy as np
import pickle
import shutil
import json
import os

def reapplyClustering(modelUsed, nbConditions, nbCluster, outputFolderResult):
  
  nbFramesTakenIntoAccount = 24
  scaleGraphs   = 1
  showFigures   = 0
  nbPCAComp     = -1 #10 # Default: -1: all components
  
  with open(os.path.join(outputFolderResult, 'boutParameters.pkl'), 'rb') as pickle_file:
    data = pickle.load(pickle_file)
  
  pca_result = np.array(data['pca_result'])
  dfParam    = data['dfParam']
  
  if nbPCAComp != -1:
    pca_result = pca_result[:, 0:nbPCAComp]
  
  if False:
    from matplotlib.pyplot import hist2d
    from scipy.stats import chi2
    hist2d(pca_result[:, 0], pca_result[:, 1], bins=500)
    plt.show()
  
  # Learning (and predicting) first model
  
  if modelUsed == 'KMeans':
    
    model = KMeans(n_clusters = nbCluster)
  
  elif modelUsed == 'AffinityPropagation':
  
    model = cluster.AffinityPropagation(damping=0.9, preference=-200, random_state=0)  
  
  elif modelUsed == 'MeanShift':
    
    bandwidth = cluster.estimate_bandwidth(pca_result, quantile=0.3)
    model = cluster.MeanShift(bandwidth=bandwidth, bin_seeding=True)
  
  elif modelUsed == 'SpectralClustering':
    
    model = cluster.SpectralClustering(n_clusters=nbCluster, eigen_solver="arpack", affinity="nearest_neighbors")
    
  elif modelUsed == 'Ward':
    
    from sklearn.neighbors import kneighbors_graph
    connectivity = kneighbors_graph(pca_result, n_neighbors= 10, include_self=False)
    connectivity = 0.5 * (connectivity + connectivity.T)
    model = cluster.AgglomerativeClustering(n_clusters=nbCluster, linkage="ward", connectivity=connectivity)
  
  elif modelUsed == 'AgglomerativeClustering':
  
    from sklearn.neighbors import kneighbors_graph
    connectivity = kneighbors_graph(pca_result, n_neighbors= 10, include_self=False)
    connectivity = 0.5 * (connectivity + connectivity.T)
    model = cluster.AgglomerativeClustering(linkage="average", affinity="cityblock", n_clusters=nbCluster,  connectivity=connectivity)
  
  elif modelUsed == 'DBSCAN':
    
    model = cluster.DBSCAN(eps=4)  
  
  elif modelUsed == 'OPTICS':
  
    model = cluster.OPTICS(min_samples=20, xi=0.05, min_cluster_size=0.1)
    
  elif modelUsed == 'BIRCH':
    
    model = cluster.Birch(n_clusters=nbCluster)
    
  elif modelUsed == 'GaussianMixture':
    
    model = GaussianMixture(n_components = nbCluster)
  
  else:
    
    model = 0 # KMeans(n_clusters = nbCluster)
  
  if type(model) != int:
  
    model.fit(pca_result)
    
    if modelUsed == 'DBSCAN':
      labels = model.labels_
      print("Labels:", np.unique(labels))
      nbCluster = len(np.unique(labels))
      model.labels_[model.labels_ == -1] = nbCluster - 1
    else:
      labels = model.predict(pca_result)
  
  else:
    
    labels = np.zeros(len(pca_result))
    
    for i in range(0, nbCluster):
      quantileBef = np.quantile(dfParam[modelUsed], i/nbCluster)
      quantileAft = np.quantile(dfParam[modelUsed], (i+1)/nbCluster)
      print("i, quantileBef, quantileAft:", i, quantileBef, quantileAft)
      labels[np.logical_and(dfParam[modelUsed] > quantileBef, dfParam[modelUsed] <= quantileAft)] = i
      print("median", str(i), " : ", np.median(dfParam[np.logical_and(dfParam[modelUsed] > quantileBef, dfParam[modelUsed] <= quantileAft)][modelUsed]))
    
    if False:
      import seaborn as sns
      dfParamWT        = dfParam[dfParam["Condition"] == "WT"]
      dfParamParkinson = dfParam[dfParam["Condition"] == "Parkinson"]
      plt.hist(dfParamWT["maxTailAngleAmplitude"].tolist(), 100)
      plt.show()
      plt.hist(dfParamParkinson["maxTailAngleAmplitude"].tolist(), 100)
      plt.show()
      # sns.histplot(data = dfParam, y=modelUsed, hue = "Condition")
    
  
  if modelUsed == 'GaussianMixture':
    predictedProbas = model.predict_proba(pca_result)
  
  # Sorting labels
  if type(model) != int:
    nbLabels       = nbCluster
    labels2        = np.zeros(len(labels))
    nbElemPerClass = np.zeros(nbLabels) 
    for i in range(0, nbLabels):
      nbElemPerClass[i] = labels.tolist().count(i)
    sortedIndices = (-nbElemPerClass).argsort()
    
    for i in range(0, len(labels)):
      labels2[i] = np.where(sortedIndices==labels[i])[0][0]
  else:
    labels2 = labels
  
  dfParam['classification'] = labels2
  
  if modelUsed == 'GaussianMixture':
    for j in range(0, nbLabels):
      probasClassJ = predictedProbas[:, sortedIndices[j]]
      dfParam['classProba' + str(j)] = probasClassJ
  
  
  print("Prediction:")
  print(labels2)
  
  predictedProb = []
  
  outputFolderResult3 = os.path.join(outputFolderResult, 'reaplyClustering_' + modelUsed[0:10] + '_' + str(nbCluster))
  if os.path.exists(outputFolderResult3):
    shutil.rmtree(outputFolderResult3)
  os.mkdir(outputFolderResult3)
  
  visualizeClusters(dfParam, labels2, predictedProb, modelUsed, nbConditions, nbCluster, nbFramesTakenIntoAccount, scaleGraphs, showFigures, outputFolderResult3, 0, 1)
  
  ###
  
  print("Silhouette Score:")
  silhouetteScore = metrics.silhouette_score(pca_result, labels2, metric='euclidean')
  print(silhouetteScore)
  print("")
  print("Calinski Harabasz Score:")
  calinskiHarabaszScore = metrics.calinski_harabasz_score(pca_result, labels2)
  print(calinskiHarabaszScore)
  print("")
  print("Davies Bouldin Score:")
  daviesBouldinScore = metrics.davies_bouldin_score(pca_result, labels2)
  print(daviesBouldinScore)
  
  if modelUsed == 'KMeans':
    elbowMethod = model.inertia_
  else:
    elbowMethod = -1
  
  return [silhouetteScore, calinskiHarabaszScore, daviesBouldinScore, elbowMethod]